from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import ResultsQuery, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetPrefixesWithAmplification(ResultsQuery):
    """
    Return the prefixes for which we have more than one reply per (flow ID, TTL).

    .. important:: This query assumes that a single probe is sent per (flow ID, TTL) pair.

    >>> from diamond_miner.test import addr_to_string, url
    >>> rows = GetPrefixesWithAmplification().execute(url, "test_invalid_prefixes")
    >>> [addr_to_string(x[2]) for x in rows]
    ['201.0.0.0', '202.0.0.0']
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT DISTINCT
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            -- This column is to simplify the InsertPrefixes query.
            1 AS has_amplification
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            probe_dst_addr,
            probe_src_port,
            probe_dst_port,
            probe_ttl
        )
        HAVING count() > 1
        """


@dataclass(frozen=True)
class GetPrefixesWithLoops(ResultsQuery):
    """
    Return the prefixes for which an IP address appears multiple time for a single flow ID.

    .. note:: Prefixes with amplification (multiple replies per probe) may trigger a false positive
              for this query, since we do not check that the IP appears at two *different* TTLs.

    >>> from diamond_miner.test import addr_to_string, url
    >>> rows = GetPrefixesWithLoops().execute(url, "test_invalid_prefixes")
    >>> [addr_to_string(x[2]) for x in rows]
    ['201.0.0.0']
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT DISTINCT
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            -- This column is to simplify the InsertPrefixes query.
            1 AS has_loops
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            probe_dst_addr,
            probe_src_port,
            probe_dst_port
        )
        HAVING uniqExact(reply_src_addr) < count()
        """



from dataclasses import asdict, dataclass
from typing import Sequence

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import (
    CreateFlowsView,
    CreateLinksTable,
    CreatePrefixesTable,
    CreateResultsTable,
)
from diamond_miner.queries.query import Query
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreateTables(Query):
    """Create the tables necessary for a measurement."""

    def statements(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> Sequence[str]:
        return (
            *CreateResultsTable(**asdict(self)).statements(measurement_id, subset),
            *CreateFlowsView(**asdict(self)).statements(measurement_id, subset),
            *CreateLinksTable(**asdict(self)).statements(measurement_id, subset),
            *CreatePrefixesTable(**asdict(self)).statements(measurement_id, subset),
        )



from .count import Count
from .count_rows import CountLinksPerPrefix, CountResultsPerPrefix
from .create_flows_view import CreateFlowsView
from .create_links_table import CreateLinksTable
from .create_prefixes_table import CreatePrefixesTable
from .create_results_table import CreateResultsTable
from .create_tables import CreateTables
from .drop_tables import DropTables
from .get_invalid_prefixes import GetPrefixesWithAmplification, GetPrefixesWithLoops
from .get_links import GetLinks
from .get_links_from_view import GetLinksFromView
from .get_next_round import GetNextRound
from .get_nodes import GetNodes
from .get_prefixes import GetPrefixes
from .get_results import GetResults
from .get_sliding_prefixes import GetSlidingPrefixes
from .insert_links import InsertLinks
from .insert_prefixes import InsertPrefixes
from .query import (
    AddrType,
    FlowsQuery,
    LinksQuery,
    PrefixesQuery,
    Query,
    ResultsQuery,
    flows_table,
    links_table,
    prefixes_table,
    results_table,
)

__all__ = (
    "AddrType",
    "Count",
    "CountLinksPerPrefix",
    "CountResultsPerPrefix",
    "CreateFlowsView",
    "CreateLinksTable",
    "CreatePrefixesTable",
    "CreateResultsTable",
    "CreateTables",
    "DropTables",
    "GetLinks",
    "GetLinksFromView",
    "GetNextRound",
    "GetNodes",
    "GetPrefixes",
    "GetResults",
    "GetSlidingPrefixes",
    "GetPrefixesWithAmplification",
    "GetPrefixesWithLoops",
    "InsertLinks",
    "InsertPrefixes",
    "Query",
    "FlowsQuery",
    "LinksQuery",
    "PrefixesQuery",
    "ResultsQuery",
    "flows_table",
    "links_table",
    "prefixes_table",
    "results_table",
)



"""
Functions for mapping flow IDs to addresses and ports.
We make the flow ID start at 0.
``prefix_size`` is the number of addresses in the prefix:
``2 ** (32 - 24)`` for a /24 in IPv4.
"""
import random

from libc.stdint cimport uint8_t, uint16_t

from pygfc import Permutation

from diamond_miner.defaults import DEFAULT_PREFIX_SIZE_V4


cdef extern from *:
    ctypedef long long int128_t "__int128_t"
    ctypedef unsigned long long uint128_t "__uint128_t"


cdef class SequentialFlowMapper:
    """
    Maps flow 0 to address 0, flow 1 to address 1, and so on until we have done
    the whole prefix. It then increases the port number in the same manner.
    """

    cdef uint128_t prefix_size

    def __init__(self, uint128_t prefix_size = DEFAULT_PREFIX_SIZE_V4):
        assert prefix_size > 0, "prefix_size must be positive."
        self.prefix_size = prefix_size

    cpdef uint128_t flow_id(self, uint128_t addr_offset, uint16_t port_offset, uint128_t prefix = 0):
        return addr_offset + port_offset

    cpdef (uint128_t, uint16_t) offset(self, uint128_t flow_id, uint128_t prefix = 0):
        if flow_id < self.prefix_size:
            return flow_id, 0
        return self.prefix_size - 1, flow_id - self.prefix_size + 1


cdef class IntervalFlowMapper:
    """
    Similar to the :py:class:`SequentialFlowMapper` but with an increment >= 1.
    This allows to target addresses .1, .33, .65, ... in priority,
    which are more likely to respond to probes :cite:`fan2010selecting`.
    """

    cdef uint128_t period
    cdef uint128_t prefix_size
    cdef uint128_t step

    def __init__(self, uint128_t prefix_size = DEFAULT_PREFIX_SIZE_V4, uint128_t step = 32):
        assert prefix_size > 0, "prefix_size must be positive."
        assert prefix_size % 2 == 0, "prefix_size must be pair."
        assert step > 0, "step must be positive."
        assert step % 2 == 0, "step must be pair."
        self.period = prefix_size // step
        self.prefix_size = prefix_size
        self.step = step

    cpdef uint128_t flow_id(self, uint128_t addr_offset, uint16_t port_offset, uint128_t prefix = 0):
        if addr_offset == 0:
            return self.prefix_size - 1
        if port_offset != 0:
            return self.prefix_size + port_offset - 1
        q, r = divmod(addr_offset - 1, self.step)
        return r * self.period + q

    cpdef (uint128_t, uint16_t) offset(self, uint128_t flow_id, uint128_t prefix = 0):
        if flow_id < self.prefix_size - 1:
            return ((flow_id * self.step) % (self.prefix_size - 1)) + 1, 0
        if flow_id == self.prefix_size - 1:
            return 0, 0
        return self.prefix_size - 1, flow_id - self.prefix_size + 1


cdef class ReverseByteFlowMapper:
    """
    Maps flow ``n`` to address ``reverse(n)`` until we have done the whole prefix.
    It then increases the port number sequentially.
    """

    cpdef uint128_t flow_id(self, uint128_t addr_offset, uint16_t port_offset, uint128_t prefix = 0):
        assert addr_offset < 256
        if addr_offset == 0:
            return 255
        if port_offset != 0:
            return 255 + port_offset
        return self.reverse_byte(addr_offset - 1)

    cpdef (uint128_t, uint16_t) offset(self, uint128_t flow_id, uint128_t prefix = 0):
        if flow_id < 255:
            return self.reverse_byte(flow_id) + 1, 0
        if flow_id == 255:
            return 0, 0
        return 255, flow_id - 255

    cdef uint8_t reverse_byte(self, uint8_t i):
        # https://stackoverflow.com/a/2602885
        i = (i & 0xF0) >> 4 | (i & 0x0F) << 4
        i = (i & 0xCC) >> 2 | (i & 0x33) << 2
        i = (i & 0xAA) >> 1 | (i & 0x55) << 1
        return i


cdef class RandomFlowMapper:
    """
    Similar to the :py:class:`SequentialFlowMapper` but with a random mapping
    between flow IDs and addresses.
    The mapping is randomized by prefix.
    """

    cdef list permutations
    cdef uint128_t prefix_size

    def __init__(self, int seed, uint128_t prefix_size = DEFAULT_PREFIX_SIZE_V4):
        # We can generate a random permutation up to 2^64-1 only.
        assert prefix_size > 0, "prefix_size must be positive."
        self.permutations = []
        self.prefix_size = min(prefix_size, (2 ** 64) - 1)
        random.seed(seed)
        for i in range(1024):
            perm = Permutation(self.prefix_size, 3, random.randint(0, 2 ** 64))
            self.permutations.append(perm)

    cpdef uint128_t flow_id(self, uint128_t addr_offset, uint16_t port_offset, uint128_t prefix):
        assert addr_offset < self.prefix_size
        if port_offset != 0:
            return self.prefix_size + port_offset - 1
        perm = self.permutations[prefix % len(self.permutations)]
        return perm.inv(addr_offset)

    cpdef (uint128_t, uint16_t) offset(self, uint128_t flow_id, uint128_t prefix):
        if flow_id < self.prefix_size:
            perm = self.permutations[prefix % len(self.permutations)]
            return perm[flow_id], 0
        else:
            return self.prefix_size - 1, flow_id - self.prefix_size + 1



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import Query, prefixes_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreatePrefixesTable(Query):
    """Create the table containing (invalid) prefixes."""

    SORTING_KEY = "probe_protocol, probe_src_addr, probe_dst_prefix"

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        CREATE TABLE IF NOT EXISTS {prefixes_table(measurement_id)}
        (
            probe_protocol         UInt8,
            probe_src_addr         IPv6,
            probe_dst_prefix       IPv6,
            has_amplification      UInt8,
            has_loops              UInt8
        )
        ENGINE MergeTree
        ORDER BY ({self.SORTING_KEY})
        """



from ipaddress import ip_network
from typing import Any, Iterable, Optional

from diamond_miner.typing import IPNetwork


def cut_ipv6(column: str, prefix_len_v4: int, prefix_len_v6: int) -> str:
    """
    >>> cut_ipv6("col", 24, 64)
    'toIPv6(cutIPv6(col, 8, 1))'
    """
    bytes_v4 = int(4 - (prefix_len_v4 / 8))
    bytes_v6 = int(16 - (prefix_len_v6 / 8))
    return f"toIPv6(cutIPv6({column}, {bytes_v6}, {bytes_v4}))"


def eq(column: str, value: Optional[Any]) -> str:
    """
    >>> eq("col", None)
    '1'
    >>> eq("col", 1)
    'col = 1'
    >>> eq("col", "1")
    "col = '1'"
    """
    if not value:
        return "1"
    if isinstance(value, str):
        return f"{column} = '{value}'"
    return f"{column} = {value}"


def leq(column: str, value: Optional[Any]) -> str:
    """
    >>> leq("col", None)
    '1'
    >>> leq("col", 1)
    'col <= 1'
    >>> leq("col", "1")
    "col <= '1'"
    """
    if not value:
        return "1"
    if isinstance(value, str):
        return f"{column} <= '{value}'"
    return f"{column} <= {value}"


def not_(column: str) -> str:
    """
    >>> not_("col")
    'NOT col'
    """
    return f"NOT {column}"


def in_(column: str, values: Iterable[Any]) -> str:
    """
    >>> in_("col", [])
    '1'
    >>> in_("col", [1,2,3])
    'col in [1,2,3]'
    """
    if not values:
        return "1"
    return f"{column} in [{','.join(map(str, values))}]"


def ipv6(x: Any) -> str:
    """
    >>> ipv6("8.8.8.8")
    "toIPv6('8.8.8.8')"
    """
    return f"toIPv6('{x}')"


def ip_eq(column: str, value: Optional[str]) -> str:
    """
    >>> ip_eq("col", None)
    '1'
    >>> ip_eq("col", "8.8.8.8")
    "col = toIPv6('8.8.8.8')"
    """
    if not value:
        return "1"
    return f"{column} = {ipv6(value)}"


def ip_in(column: str, subset: Optional[IPNetwork]) -> str:
    """
    >>> ip_in("col", None)
    '1'
    >>> ip_in("col", ip_network("::/0"))
    "(col >= toIPv6('::') AND col <= toIPv6('ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'))"
    """
    if not subset:
        return "1"
    return f"({column} >= {ipv6(subset[0])} AND {column} <= {ipv6(subset[-1])})"


def ip_not_in(column: str, subset: Optional[IPNetwork]) -> str:
    """
    >>> ip_not_in("col", None)
    '1'
    >>> ip_not_in("col", ip_network("::/0"))
    "(col < toIPv6('::') OR col > toIPv6('ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'))"
    """
    if not subset:
        return "1"
    return f"({column} < {ipv6(subset[0])} OR {column} > {ipv6(subset[-1])})"


def ip_not_private(column: str) -> str:
    return f"""
    {ip_not_in(column, ip_network('10.0.0.0/8'))}
    AND {ip_not_in(column, ip_network('172.16.0.0/12'))}
    AND {ip_not_in(column, ip_network('192.168.0.0/16'))}
    AND {ip_not_in(column, ip_network('fd00::/8'))}
    """


def and_(a: str, b: str) -> str:
    """
    >>> and_("0", "1")
    '(0 AND 1)'
    """
    return f"({a} AND {b})"


def or_(a: str, b: str) -> str:
    """
    >>> or_("0", "1")
    '(0 OR 1)'
    """
    return f"({a} OR {b})"



from distutils.command.build_ext import build_ext  # type: ignore

from Cython.Build import cythonize


def build(setup_kwargs):
    setup_kwargs.update(
        {
            "cmdclass": {"build_ext": build_ext},
            "ext_modules": cythonize(
                "diamond_miner/**/*.pyx",
                compiler_directives={"binding": True, "embedsignature": True},
                language_level=3,
            ),
        }
    )



from dataclasses import dataclass

from diamond_miner.defaults import (
    DEFAULT_PREFIX_LEN_V4,
    DEFAULT_PREFIX_LEN_V6,
    UNIVERSE_SUBSET,
)
from diamond_miner.queries.fragments import cut_ipv6
from diamond_miner.queries.query import Query, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreateResultsTable(Query):
    """Create the table used to store the measurement results from the prober."""

    SORTING_KEY = "probe_protocol, probe_src_addr, probe_dst_prefix, probe_dst_addr, probe_src_port, probe_dst_port, probe_ttl"

    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        CREATE TABLE IF NOT EXISTS {results_table(measurement_id)}
        (
            probe_protocol         UInt8,
            probe_src_addr         IPv6,
            probe_dst_addr         IPv6,
            probe_src_port         UInt16,
            probe_dst_port         UInt16,
            probe_ttl              UInt8,
            quoted_ttl             UInt8,
            reply_src_addr         IPv6,
            reply_protocol         UInt8,
            reply_icmp_type        UInt8,
            reply_icmp_code        UInt8,
            reply_ttl              UInt8,
            reply_size             UInt16,
            reply_mpls_labels      Array(UInt32),
            -- The rtt column is the largest compressed column, we use T64 and ZSTD to reduce its size, see:
            -- https://altinity.com/blog/2019/7/new-encodings-to-improve-clickhouse
            -- https://clickhouse.tech/docs/en/sql-reference/statements/create/table/#codecs
            rtt                    UInt16 CODEC(T64, ZSTD(1)),
            round                  UInt8,
            -- Materialized columns
            probe_dst_prefix       IPv6 MATERIALIZED {cut_ipv6('probe_dst_addr', self.prefix_len_v4, self.prefix_len_v6)},
            reply_src_prefix       IPv6 MATERIALIZED {cut_ipv6('reply_src_addr', self.prefix_len_v4, self.prefix_len_v6)},
            -- https://en.wikipedia.org/wiki/Reserved_IP_addresses
            private_probe_dst_prefix UInt8 MATERIALIZED
                (probe_dst_prefix >= toIPv6('0.0.0.0')      AND probe_dst_prefix <= toIPv6('0.255.255.255'))   OR
                (probe_dst_prefix >= toIPv6('10.0.0.0')     AND probe_dst_prefix <= toIPv6('10.255.255.255'))  OR
                (probe_dst_prefix >= toIPv6('100.64.0.0')   AND probe_dst_prefix <= toIPv6('100.127.255.255')) OR
                (probe_dst_prefix >= toIPv6('127.0.0.0')    AND probe_dst_prefix <= toIPv6('127.255.255.255')) OR
                (probe_dst_prefix >= toIPv6('172.16.0.0')   AND probe_dst_prefix <= toIPv6('172.31.255.255'))  OR
                (probe_dst_prefix >= toIPv6('192.0.0.0')    AND probe_dst_prefix <= toIPv6('192.0.0.255'))     OR
                (probe_dst_prefix >= toIPv6('192.0.2.0')    AND probe_dst_prefix <= toIPv6('192.0.2.255'))     OR
                (probe_dst_prefix >= toIPv6('192.88.99.0')  AND probe_dst_prefix <= toIPv6('192.88.99.255'))   OR
                (probe_dst_prefix >= toIPv6('192.168.0.0')  AND probe_dst_prefix <= toIPv6('192.168.255.255')) OR
                (probe_dst_prefix >= toIPv6('198.18.0.0')   AND probe_dst_prefix <= toIPv6('198.19.255.255'))  OR
                (probe_dst_prefix >= toIPv6('198.51.100.0') AND probe_dst_prefix <= toIPv6('198.51.100.255'))  OR
                (probe_dst_prefix >= toIPv6('203.0.113.0')  AND probe_dst_prefix <= toIPv6('203.0.113.255'))   OR
                (probe_dst_prefix >= toIPv6('224.0.0.0')    AND probe_dst_prefix <= toIPv6('239.255.255.255')) OR
                (probe_dst_prefix >= toIPv6('233.252.0.0')  AND probe_dst_prefix <= toIPv6('233.252.0.255'))   OR
                (probe_dst_prefix >= toIPv6('240.0.0.0')    AND probe_dst_prefix <= toIPv6('255.255.255.255')) OR
                (probe_dst_prefix >= toIPv6('fd00::')       AND probe_dst_prefix <= toIPv6('fdff:ffff:ffff:ffff:ffff:ffff:ffff:ffff')),
            private_reply_src_addr UInt8 MATERIALIZED
                (reply_src_addr >= toIPv6('0.0.0.0')        AND reply_src_addr <= toIPv6('0.255.255.255'))     OR
                (reply_src_addr >= toIPv6('10.0.0.0')       AND reply_src_addr <= toIPv6('10.255.255.255'))    OR
                (reply_src_addr >= toIPv6('100.64.0.0')     AND reply_src_addr <= toIPv6('100.127.255.255'))   OR
                (reply_src_addr >= toIPv6('127.0.0.0')      AND reply_src_addr <= toIPv6('127.255.255.255'))   OR
                (reply_src_addr >= toIPv6('172.16.0.0')     AND reply_src_addr <= toIPv6('172.31.255.255'))    OR
                (reply_src_addr >= toIPv6('192.0.0.0')      AND reply_src_addr <= toIPv6('192.0.0.255'))       OR
                (reply_src_addr >= toIPv6('192.0.2.0')      AND reply_src_addr <= toIPv6('192.0.2.255'))       OR
                (reply_src_addr >= toIPv6('192.88.99.0')    AND reply_src_addr <= toIPv6('192.88.99.255'))     OR
                (reply_src_addr >= toIPv6('192.168.0.0')    AND reply_src_addr <= toIPv6('192.168.255.255'))   OR
                (reply_src_addr >= toIPv6('198.18.0.0')     AND reply_src_addr <= toIPv6('198.19.255.255'))    OR
                (reply_src_addr >= toIPv6('198.51.100.0')   AND reply_src_addr <= toIPv6('198.51.100.255'))    OR
                (reply_src_addr >= toIPv6('203.0.113.0')    AND reply_src_addr <= toIPv6('203.0.113.255'))     OR
                (reply_src_addr >= toIPv6('224.0.0.0')      AND reply_src_addr <= toIPv6('239.255.255.255'))   OR
                (reply_src_addr >= toIPv6('233.252.0.0')    AND reply_src_addr <= toIPv6('233.252.0.255'))     OR
                (reply_src_addr >= toIPv6('240.0.0.0')      AND reply_src_addr <= toIPv6('255.255.255.255'))   OR
                (reply_src_addr >= toIPv6('fd00::')         AND reply_src_addr <= toIPv6('fdff:ffff:ffff:ffff:ffff:ffff:ffff:ffff')),
            destination_host_reply   UInt8 MATERIALIZED probe_dst_addr = reply_src_addr,
            destination_prefix_reply UInt8 MATERIALIZED probe_dst_prefix = reply_src_prefix,
            -- ICMP: protocol 1, UDP: protocol 17, ICMPv6: protocol 58
            valid_probe_protocol   UInt8 MATERIALIZED probe_protocol IN [1, 17, 58],
            time_exceeded_reply    UInt8 MATERIALIZED (reply_protocol = 1 AND reply_icmp_type = 11) OR (reply_protocol = 58 AND reply_icmp_type = 3)
        )
        ENGINE MergeTree
        ORDER BY ({self.SORTING_KEY})
        """



from ipaddress import IPv6Address

url = "clickhouse://localhost"


def addr_to_string(addr: IPv6Address) -> str:
    """
    >>> from ipaddress import ip_address
    >>> addr_to_string(ip_address('::dead:beef'))
    '::dead:beef'
    >>> addr_to_string(ip_address('::ffff:8.8.8.8'))
    '8.8.8.8'
    """
    return str(addr.ipv4_mapped or addr)



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import CreateFlowsView
from diamond_miner.queries.fragments import ip_in
from diamond_miner.queries.query import FlowsQuery, flows_table, prefixes_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetLinksFromView(FlowsQuery):
    """
    Compute the links from the flows view.
    This returns one line per ``(flow, link)`` pair.

    We do not emit a link in the case of single reply in a traceroute.
    For example: ``* * node * *``, does not generate a link.
    However, ``* * node * * node'``, will generate ``(node, *)`` and ``(*, node')``.

    We emit cross-rounds links.
    For example if flow N sees node A at TTL 10 at round 1 and flow N sees node B at TTL 11 at round 2,
    we will emit ``(1, 10, A) - (2, 11, B)``.

    We assume that there exists a single (flow, ttl) pair over all rounds.
    TODO: Assert this?

    If ``round_eq`` is none, compute the links per flow, across all rounds.
    Otherwise, compute the links per flow, for the specified round.
    This is useful if you want to update a `links` table round-by-round:
    such a table will contain only intra-round links but can be updated incrementally.

    >>> from diamond_miner.test import url
    >>> links = GetLinksFromView().execute(url, "test_nsdi_example")
    >>> len(links)
    58
    """

    ignore_invalid_prefixes: bool = True

    optimize_aggregation_in_order: bool = False
    """
    Necessary to avoid excessive memory usage when inserting all the links in one batch.
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        if self.ignore_invalid_prefixes:
            invalid_filter = f"""
            AND (probe_protocol, probe_src_addr, probe_dst_prefix)
            NOT IN (
                SELECT probe_protocol, probe_src_addr, probe_dst_prefix
                FROM {prefixes_table(measurement_id)}
                WHERE {ip_in('probe_dst_prefix', subset)}
                AND has_amplification
            )
            """
            # TODO: We currently do not drop prefixes with loops as this considerably
            # reduces the number of discoveries. As such, we send more probe than necessary.
            # A better way would be to detect the min/max TTL of a loop, and to ignore it
            # in the next round query.
        else:
            invalid_filter = ""

        if self.optimize_aggregation_in_order:
            optimize_fragment = "SETTINGS optimize_aggregation_in_order = 1"
        else:
            optimize_fragment = ""

        return f"""
        WITH
            -- (round, ttl, addr)
            groupUniqArrayMerge(replies) AS traceroute,
            arrayMap(x -> x.2, traceroute) AS ttls,
            arrayMap(x -> (x.1, x.3), traceroute) AS val,
            CAST((ttls, val), 'Map(UInt8, Tuple(UInt8, IPv6))') AS map,
            arrayMin(ttls) AS first_ttl,
            arrayMax(ttls) AS last_ttl,
            arrayMap(i -> (toUInt8(i), toUInt8(i + 1), map[toUInt8(i)], map[toUInt8(i + 1)]), range(first_ttl, last_ttl)) AS links,
            arrayJoin(links) AS link
        SELECT
            {CreateFlowsView.SORTING_KEY},
            -- Set the round number for partial links:
            -- The link (1, 10, A) -> (null, 11, *) becomes
            --          (1, 10, A) -> (1,    11, *)
            if(link.3.1 != 0, link.3.1, link.4.1) AS near_round,
            if(link.4.1 != 0, link.4.1, link.3.1) AS far_round,
            link.1 AS near_ttl,
            link.2 AS far_ttl,
            link.3.2 AS near_addr,
            link.4.2 AS far_addr
        FROM {flows_table(measurement_id)}
        WHERE {self.filters(subset)}
        {invalid_filter}
        GROUP BY ({CreateFlowsView.SORTING_KEY})
        {optimize_fragment}
        """



from ipaddress import IPv6Network

DEFAULT_PREFIX_LEN_V4 = 24
"""Default prefix length for IPv4."""
DEFAULT_PREFIX_LEN_V6 = 64
"""Default prefix length for IPv6."""

DEFAULT_PREFIX_SIZE_V4 = 2 ** (32 - DEFAULT_PREFIX_LEN_V4)
"""Default prefix size (number of addresses) for IPv4."""
DEFAULT_PREFIX_SIZE_V6 = 2 ** (128 - DEFAULT_PREFIX_LEN_V6)
"""Default prefix size (number of addresses) for IPv6."""

DEFAULT_PROBE_SRC_PORT = 24000
"""Default probe source port. Encoded in the ICMP checksum field for ICMP probes."""
DEFAULT_PROBE_DST_PORT = 33434
"""Default probe destination port. Unused for ICMP probes."""

UNIVERSE_SUBSET = IPv6Network("::/0")
"""Set of all possible IP addresses."""

PROTOCOLS = {1: "icmp", 17: "udp", 58: "icmp6"}
"""Mapping of IP protocol numbers to caracal protocol strings."""



from ipaddress import IPv4Network, IPv6Network
from typing import Protocol, Tuple, Union


class FlowMapper(Protocol):
    """Protocol to which a flow mapper must conform."""

    def flow_id(self, addr_offset: int, port_offset: int, prefix: int) -> int:
        """Return the flow ID for a given address and port offset."""
        ...

    def offset(self, flow_id: int, prefix: int) -> Tuple[int, int]:
        """Return the address and port offset for a given flow ID."""
        ...


Probe = Tuple[int, int, int, int, str]
IPNetwork = Union[IPv4Network, IPv6Network]



from random import randint

from pygfc import Permutation


# The Cython version is approx. 2x faster on a M1 CPU.
cdef class ParameterGrid:
    cdef tuple parameters
    cdef tuple size_
    cdef len

    def __init__(self, *parameters):
        self.parameters = parameters
        self.size_ = tuple(len(p) for p in self.parameters)
        self.len = 1
        for dim in self.size_:
            self.len *= dim

    def __getitem__(self, index):
        if isinstance(index, int):
            if index >= len(self):
                raise IndexError("index out of range")
            index = self.linear_to_subscript(index)
        return [p[i] for p, i in zip(self.parameters, index)]

    def __iter__(self):
        return (self[index] for index in range(len(self)))

    def __len__(self):
        return self.len

    @property
    def size(self):
        return self.size_

    def shuffled(self, int rounds = 6, seed = None):
        seed = seed or randint(0, 2 ** 64 - 1)
        perm = Permutation(len(self), rounds, seed)
        return (self[index] for index in perm)

    cdef list linear_to_subscript(self, index):
        cdef list coordinates = []
        for dim in self.size_:
            index, coordinate = divmod(index, dim)
            coordinates.append(coordinate)
        return coordinates



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['diamond_miner', 'diamond_miner.queries', 'diamond_miner.rounds']

package_data = \
{'': ['*']}

install_requires = \
['aioch>=0.0.2,<0.0.3',
 'clickhouse-driver>=0.2.0,<0.3.0',
 'pygfc>=1.0.3,<2.0.0',
 'zstandard>=0.15.2,<0.16.0']

setup_kwargs = {
    'name': 'diamond-miner',
    'version': '0.1.4',
    'description': 'High-speed, Internet-scale, load-balanced paths discovery.',
    'long_description': '# Diamond-Miner 💎\n\n[![Tests](https://img.shields.io/github/workflow/status/dioptra-io/diamond-miner/Tests?logo=github)](https://github.com/dioptra-io/diamond-miner/actions/workflows/tests.yml)\n[![Coverage](https://img.shields.io/codecov/c/github/dioptra-io/diamond-miner?logo=codecov&logoColor=white&token=RKZSQ2CL4J)](https://app.codecov.io/gh/dioptra-io/diamond-miner)\n[![Documentation](https://img.shields.io/badge/documentation-online-blue.svg?logo=read-the-docs&logoColor=white)](https://dioptra-io.github.io/diamond-miner/)\n[![PyPI](https://img.shields.io/pypi/v/diamond-miner?logo=pypi&logoColor=white)](https://pypi.org/project/diamond-miner/)\n\n> D-Miner is the first Internet-scale system that captures a multipath view of the topology.\n> By combining and adapting state-of-the-art multipath detection and high speed randomized topology discovery techniques,\n> D-Miner permits discovery of the Internet’s multipath topology in 2.5 days¹ when probing at 100kpps. [[1]](#references)\n\n¹ As of v0.1.0, diamond-miner can discover the multipath topology in less than a day when probing at 100k pps.\n\n## 🚀 Quickstart\n\nDiamond-Miner is implemented as a Python library which computes the probes to send.\nTo actually send the probes, we recommend [Caracal](https://github.com/dioptra-io/caracal) but you can use the tool of your liking, such as [yarrp](https://github.com/cmand/yarrp) or [scamper](https://www.caida.org/catalog/software/scamper/) as long as you convert its output to the expected format described in the documentation.\n\nTo get started, install Diamond-Miner and head over to the [documentation](https://dioptra-io.github.io/diamond-miner/):\n```bash\npip install diamond-miner\n```\n\n## Publications\n\nDiamond-Miner has been presented and published at [NSDI 2020](https://www.usenix.org/conference/nsdi20/presentation/vermeulen).\nSince then, the code has been refactored and separated in the [`diamond-miner`](https://github.com/dioptra-io/diamond-miner) and [`caracal`](https://github.com/dioptra-io/caracal) repositories.\nThe code as it was at the time of the publication is available in the [`diamond-miner-cpp`](https://github.com/dioptra-io/diamond-miner-cpp) and [`diamond-miner-wrapper`](https://github.com/dioptra-io/diamond-miner-wrapper) repositories.\n\nIf you use Diamond-Miner, please cite the following paper:\n```bibtex\n@inproceedings {DiamondMiner2020,\n  author = {Kevin Vermeulen and Justin P. Rohrer and Robert Beverly and Olivier Fourmaux and Timur Friedman},\n  title = {Diamond-Miner: Comprehensive Discovery of the Internet{\\textquoteright}s Topology Diamonds },\n  booktitle = {17th {USENIX} Symposium on Networked Systems Design and Implementation ({NSDI} 20)},\n  year = {2020},\n  isbn = {978-1-939133-13-7},\n  address = {Santa Clara, CA},\n  pages = {479--493},\n  url = {https://www.usenix.org/conference/nsdi20/presentation/vermeulen},\n  publisher = {{USENIX} Association},\n  month = feb,\n}\n```\n\n## Authors\n\nDiamond-Miner is developed and maintained by the [Dioptra group](https://dioptra.io) at Sorbonne Université in Paris, France.\nThe initial version has been written by [Kévin Vermeulen](https://github.com/kvermeul), with subsequents refactoring and improvements by [Maxime Mouchet](https://github.com/maxmouchet) and [Matthieu Gouel](https://github.com/matthieugouel).\n\n## References\n\n[1] Vermeulen, Kevin, et al. ["Diamond-Miner: Comprehensive Discovery of the Internet\'s Topology Diamonds."](https://www.usenix.org/system/files/nsdi20-paper-vermeulen.pdf) _17th USENIX Symposium on Networked Systems Design and Implementation (NSDI 20)_. 2020.\n\n## License & Dependencies\n\nThis software is released under the [MIT license](/LICENSE), in accordance with the license of its dependencies.\n\nName                                             | License                                    | Usage\n-------------------------------------------------|--------------------------------------------|------\n[aioch](https://github.com/mymarilyn/aioch)      | [MIT](https://opensource.org/licenses/MIT) | Interacting with ClickHouse\n[clickhouse-driver](https://github.com/mymarilyn/clickhouse-driver)      | [MIT](https://opensource.org/licenses/MIT) | Interacting with ClickHouse\n[pygfc](https://github.com/maxmouchet/gfc)       | [MIT](https://opensource.org/licenses/MIT) | Generating random permutations\n[python-zstandard](https://github.com/indygreg/python-zstandard) | [3-clause BSD](https://opensource.org/licenses/BSD-3-Clause) | Compression\n',
    'author': 'Kevin Vermeulen',
    'author_email': 'kevin.vermeulen@columbia.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dioptra-io/diamond-miner',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}
from build_ext import *
build(setup_kwargs)

setup(**setup_kwargs)



from libc.stdint cimport uint8_t, uint16_t
from libc.stdio cimport snprintf
from libc.stdlib cimport free, malloc


# https://pubs.opengroup.org/onlinepubs/9699919799/basedefs/netinet_in.h.html
# > INET6_ADDRSTRLEN: 46. Length of the string form for IPv6.
cdef size_t INET6_ADDRSTRLEN = 46

def format_probe(dst_addr_v6, uint16_t src_port, uint16_t dst_port, uint8_t ttl, str protocol):
    return f"{format_ipv6(dst_addr_v6)},{src_port},{dst_port},{ttl},{protocol}"

def format_ipv6(int_addr):
    """
    Convert IPv6 uint128 to string.
    Faster than building an ip_address object and calling str().
    """
    cdef size_t size = INET6_ADDRSTRLEN + 1
    cdef char *c_str = <char *> malloc(size * sizeof(char))
    cdef int length = format_ipv6_(int_addr, c_str, size)
    if length < 0:
        return None
    py_str = c_str[:length].decode('UTF-8')
    free(c_str)
    return py_str

cdef inline int format_ipv6_(int_addr, char* c_str, size_t size):
    cdef uint16_t a, b, c, d, e, f, g, h
    a = (int_addr & 0xFFFF0000000000000000000000000000) >> 112
    b = (int_addr & 0x0000FFFF000000000000000000000000) >> 96
    c = (int_addr & 0x00000000FFFF00000000000000000000) >> 80
    d = (int_addr & 0x000000000000FFFF0000000000000000) >> 64
    e = (int_addr & 0x0000000000000000FFFF000000000000) >> 48
    f = (int_addr & 0x00000000000000000000FFFF00000000) >> 32
    g = (int_addr & 0x000000000000000000000000FFFF0000) >> 16
    h = int_addr & 0x0000000000000000000000000000FFFF
    return snprintf(c_str, size, "%X:%X:%X:%X:%X:%X:%X:%X", a, b, c, d, e, f, g, h)



from collections import namedtuple
from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import LinksQuery, links_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetNextRound(LinksQuery):
    # This query is tested in test_queries.py due to its complexity.

    adaptive_eps: bool = True
    dminer_lite: bool = True
    target_epsilon: float = 0.05

    Row = namedtuple(
        "Row",
        "protocol,dst_prefix,already_sent,to_send,ttls",
    )

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        if self.adaptive_eps:
            eps_fragment = """
            if(max_links_curr == 0, target_epsilon, 1 - exp(log(1 - target_epsilon) / max_links_curr))
                AS epsilon_curr,
            if(max_links_prev == 0, target_epsilon, 1 - exp(log(1 - target_epsilon) / max_links_prev))
                AS epsilon_prev,
            """
        else:
            eps_fragment = """
            target_epsilon AS epsilon_curr,
            target_epsilon AS epsilon_prev,
            """

        if self.dminer_lite:
            dm_fragment = """
            arrayMap(k -> toUInt32(ceil(ln(epsilon_curr / (k + 1)) / ln((k + 1 - 1) / (k + 1)))), links_per_ttl_curr) AS mda_flows_curr,
            arrayMap(k -> toUInt32(ceil(ln(epsilon_prev / (k + 1)) / ln((k + 1 - 1) / (k + 1)))), links_per_ttl_prev) AS mda_flows_prev,
            """
        else:
            # TODO: Implement by computing Dh(v)
            raise NotImplementedError

        return f"""
        WITH
            {self.round_leq} AS current_round,
            {self.target_epsilon} AS target_epsilon,
            -- 1) Compute the links
            --  x.1       x.2        x.3
            -- (near_ttl, near_addr, far_addr)
            groupUniqArray((near_ttl, near_addr, far_addr)) AS links_curr,
            groupUniqArrayIf((near_ttl, near_addr, far_addr), near_round < current_round) AS links_prev,
            -- 2) Count the number of links per TTL
            -- extract only the TTLs, this greatly speeds-up arrayCount
            arrayMap(x -> x.1, links_curr) AS TTLs_curr,
            arrayMap(x -> x.1, links_prev) AS TTLs_prev,
            -- find the min/max TTLs
            -- we add +2 since range() is exclusive and that we compute the max over the *near* TTL
            range(arrayMin(TTLs_curr), arrayMax(TTLs_curr) + 2) AS TTLs,
            -- count distinct links per TTL
            arrayMap(t -> countEqual(TTLs_curr, t), TTLs) AS links_per_ttl_curr,
            arrayMap(t -> countEqual(TTLs_prev, t), TTLs) AS links_per_ttl_prev,
            -- find the maximum number of links over all TTLs
            arrayMax(links_per_ttl_curr) AS max_links_curr,
            arrayMax(links_per_ttl_prev) AS max_links_prev,
            -- 3) Determine if the prefix can be skipped at the next round
            -- 1 if no new links and nodes have been discovered in the current round
            equals(links_per_ttl_curr, links_per_ttl_prev) AS skip_prefix,
            -- 4) Compute MDA stopping points
            {eps_fragment}
            -- 5) Compute the number of probes to send during the next round
            {dm_fragment}
            -- compute the number of probes sent at previous round
            -- if round = 1: we known that we sent 6 probes
            -- if round > 1 and TTL > 1: we take the max of the DMiner formula
            -- over the links discovered during the previous rounds at TTL t and t-1
            if(
                current_round == 1,
                arrayMap(t -> 6, TTLs),
                arrayMap(i -> arrayMax([mda_flows_prev[i], mda_flows_prev[i - 1]]), arrayEnumerate(TTLs))
            ) AS already_sent,
            -- compute the number of probes to send during the next round
            -- => max of probes to send over TTL t and t-1
            arrayMap(i -> arrayMax([
                0,
                mda_flows_curr[i] - already_sent[i],
                mda_flows_curr[i - 1] - already_sent[i - 1]
            ]), arrayEnumerate(TTLs)) AS to_send
            -- TODO: Cleanup/optimize/rewrite/... below
            -- do not send probes to TTLs where no replies have been received
            -- it is unlikely that we will discover more at this TTL if the first 6 flows have seen nothing
            -- arrayMap(t -> arrayUniq(arrayFilter(x -> x != toIPv6('::'), arrayMap(x -> x.2, arrayFilter(x -> x.1 == t, links_curr)))), TTLs) AS nodes_per_ttl_near,
            -- arrayMap(t -> arrayUniq(arrayFilter(x -> x != toIPv6('::'), arrayMap(x -> x.3, arrayFilter(x -> x.1 + 1 == t, links_curr)))), TTLs) AS nodes_per_ttl_far,
            -- arrayMap(t -> arrayMax([nodes_per_ttl_near[t], nodes_per_ttl_far[t]]), TTLs) AS nodes_per_ttl,
            -- arrayMap(t -> if(nodes_per_ttl[t] > 0, probes[t], 0), TTLs) AS probes_final
        SELECT
            probe_protocol,
            probe_dst_prefix,
            already_sent,
            to_send,
            TTLs
        FROM {links_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (probe_protocol, probe_src_addr, probe_dst_prefix)
        HAVING NOT skip_prefix
        """
